# Copyright (c) 2025 Tatu Aalto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import sys
from io import StringIO
from pathlib import Path

import typer
from robocop.run import format_files  # type: ignore

from .config import get_config, logger, set_robocop_config_file


class Capturing(list):
    def __enter__(self) -> "Capturing":
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = self._stringio_out = StringIO()
        sys.stderr = self._stringio_err = StringIO()
        return self

    def __exit__(self, *_: object) -> None:
        stdout_content = self._stringio_out.getvalue()
        stderr_content = self._stringio_err.getvalue()
        self.extend(stdout_content.splitlines())
        self.extend(stderr_content.splitlines())
        del self._stringio_out
        del self._stringio_err
        sys.stdout = self._stdout
        sys.stderr = self._stderr


async def robocop_format(path: Path) -> str:
    sources = [path]
    kwargs = {"sources": sources}
    config = get_config()
    kwargs = set_robocop_config_file(config, kwargs)
    raised_error = None
    with Capturing() as output:
        try:
            format_files(**kwargs)
        except typer.Exit:
            pass
        except Exception as error:  # noqa
            logger.error("Error during RoboCop format: %s", error)
            raised_error = error
    report = "\n".join(output) if not raised_error else f"Error during formatting: {raised_error}"
    logger.info("RoboCop format completed with report: %s", report)
    return report



# Copyright (c) 2025 Tatu Aalto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from pathlib import Path

from mcp.server.fastmcp import FastMCP

from .config import get_config, logger, resolve_path
from .mcp_check import filter_violations, format_report, get_violation_fix, run_robocop
from .mcp_format import robocop_format

mcp = FastMCP("op-robocop-mcp")


@mcp.tool()
async def get_robocop_report(path: str | None) -> str:
    """
    Run RoboCop on the provided source code and return the report.

    Args:
        path (str | None): The path to folder or a file to analyze. If None, uses the
        current directory for analysis.

    Returns:
        str: The Robocop report in markdown format. Report by default contains 20 first violations
        of the same type. If there are other types of violations or there are more than 20 same type
        of violations, a note is added about how many more violations were found but not shown.
        Report contains at the end also a proposed fix for the first violation.

    Example if there is one Violation in path, which looks like this:
    Violation(
        file=WindowsPath('sample.robot'),
        line_number=2,
        end_line=2,
        column=1,
        end_column=15,
        severity='W',
        rule_id='DOC02',
        description="Missing documentation in 'this is a test' test case"
    )
    Then return value would look like this:
    # Robocop Report

    ## Violation for file sample.robot in line 2 rule DOC02

    description: Missing documentation in 'this is a test' test case
    start line: 2
    end line: 2
    start column: 1
    end column: 15
    file: C:\\path\\to\\sample.robot
    rule id: DOC02
    severity: WARNING

    All violations reported.

    ## Proposed fixe for violations
    The following fix is proposed: Add documentation to the 'this is a test' test case

    """
    path_resolved = resolve_path(path)
    logger.info("Running Robocop check on path: '%s'", path_resolved)
    report = await run_robocop(path_resolved)
    filter_report = filter_violations(report)
    if not filter_report:
        logger.info("No violations found.")
        return "# Robocop Report\n\nNo violations found."
    logger.info("Dump to markdown...")
    markdown_lines = ["# Robocop Report", ""]
    for item in filter_report:
        markdown_lines.extend(format_report(item))

    first_violation = filter_report[0]
    config = get_config()
    proposed_fix = get_violation_fix(first_violation, config)
    markdown_lines.extend(
        [
            "",
            "## Proposed fixe for violations",
            "",
            f"The following fix is proposed: {proposed_fix}",
        ],
    )
    if len(filter_report) < len(report):
        markdown_lines.append(f"\nand {len(report) - len(filter_report)} more violations not shown.")
    else:
        markdown_lines.append("\nAll violations reported.")
    return "\n".join(markdown_lines)


@mcp.tool()
async def run_robocop_format(path: str | None) -> str:
    """Runs RoboCop formatter on the given path or folder.

    Args:
        path (str | None): The path to folder or a file to format. If None, uses the
        current directory for analysis.
    Returns:
        str: A summary of the operation.
    """
    path_resolved = resolve_path(path)
    logger.info("Running Robocop format on path: '%s'", path_resolved)
    return await robocop_format(Path(path_resolved))


def main() -> None:
    "Main to run the robocop-mcp server."
    logger.info("Starting OP Robocop MCP...")
    config = get_config()
    if config.robocopmcp_config_file:
        logger.info("With %s file.", config.robocopmcp_config_file)
    mcp.run(transport="stdio")


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Tatu Aalto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from pathlib import Path

from robocop.linter.diagnostics import Diagnostic  # type: ignore
from robocop.run import check_files  # type: ignore

from .config import Config, get_config, logger, set_robocop_config_file


@dataclass
class Violation:
    file: Path
    start_line: int
    end_line: int
    start_column: int
    end_column: int
    severity: str
    rule_id: str
    description: str


def _convert_to_violations(result: list[Diagnostic]) -> list[Violation]:
    logger.info("Convert to violations")
    return [
        Violation(
            file=Path(item.source),
            start_line=item.range.start.line,
            end_line=item.range.end.line,
            start_column=item.range.start.character,
            end_column=item.range.end.character,
            severity=item.severity.name,
            rule_id=item.rule.rule_id,
            description=item.message,
        )
        for item in result
    ]


async def run_robocop(path: str) -> list[Violation]:
    sources = [Path(path)]
    kwargs = {"sources": sources, "return_result": True, "silent": True}
    config = get_config()
    kwargs = set_robocop_config_file(config, kwargs)
    logger.info("Running Robocop check_files with kwargs: %s", kwargs)
    result = check_files(**kwargs)
    if result is None:
        return []
    return _convert_to_violations(result)


def _get_first_violation(violations: list[Violation], config: Config) -> Violation | None:
    for violation in violations:
        if violation.rule_id in config.rule_priority:
            return violation
    logger.info("No prioritized rule violation found.")
    for violation in violations:
        if violation.rule_id not in config.rule_ignore:
            return violation
    logger.info("No rule priority or ignore violations found, return first violation.")
    if violations:
        return violations[0]
    return None


def filter_violations(violations: list[Violation]) -> list[Violation]:
    config = get_config()
    filtered_violations: list[Violation] = []
    first_violation = _get_first_violation(violations, config)
    if first_violation is None:
        logger.info("No violations found to filter.")
        return filtered_violations
    for violation in violations:
        if len(filtered_violations) >= config.violation_count:
            break
        if violation.rule_id == first_violation.rule_id:
            filtered_violations.append(violation)
    return filtered_violations


def format_report(violation: Violation) -> list[str]:
    heading = (
        f"## Violation for file {violation.file.name} in line {violation.start_line} rule {violation.rule_id}"
    )
    return [
        heading,
        "",
        f"description: {violation.description}",
        f"start line: {violation.start_column}",
        f"end line: {violation.end_column}",
        f"start column: {violation.start_column}",
        f"end column: {violation.end_column}",
        f"file: {violation.file}",
        f"rule id: {violation.rule_id}",
        f"severity: {violation.severity}",
    ]


def _is_file(path: str) -> bool:
    try:
        return Path(path).is_file()
    except OSError:
        return False


def get_violation_fix(violation: Violation, config: Config) -> str:
    rule_id = violation.rule_id
    if rule_id in config.user_rules:
        rule_fix = config.user_rules[rule_id]
        if _is_file(rule_fix.instruction):
            with Path(rule_fix.instruction).open("r") as file:
                return file.read()
        return rule_fix.instruction
    if rule_id in config.predefined_fixes:
        rule_fix = config.predefined_fixes[rule_id]
        if _is_file(rule_fix.instruction):
            with Path(rule_fix.instruction).open("r") as file:
                return file.read()
        return rule_fix.instruction
    if rule_id in config.robocop_rules:
        rule_fix = config.robocop_rules[rule_id]
        return rule_fix.instruction
    return "No solution proposed fix found"
